"""Shim for attrs to make `Cat`s possible.

Current attrs version supported is 19.1.0

"""
# pylint: disable=redefined-builtin
import typing as ty
from decimal import Decimal

from attr._make import NOTHING, _ClassBuilder


_SCALAR_TYPES_WITH_NO_EMPTY_VALUES = (bool, float, int, Decimal)


def cat_attrs(
    maybe_cls=None,
    these=None,
    repr_ns=None,
    repr=True,
    cmp=True,
    hash=None,
    init=True,
    slots=False,
    frozen=False,
    weakref_slot=True,
    str=False,
    auto_attribs=False,
    kw_only=False,
    cache_hash=False,
    disallow_empties=True,
    auto_exc=False,
):
    """Copied from attrs._make in attrs 19.1.0 in order to allow dynamic adding of validators!

    The only difference is (or should be...) the hook_builder_before_doing_anything call.
    """

    def wrap(cls):
        if getattr(cls, "__class__", None) is None:
            raise TypeError("attrs only works with new-style classes.")

        is_exc = auto_exc is True and issubclass(cls, BaseException)

        builder = _ClassBuilder(
            cls,
            these,
            slots,
            frozen,
            weakref_slot,
            auto_attribs,
            kw_only,
            cache_hash,
            is_exc,
        )

        _hook_builder_before_doing_anything(builder, disallow_empties=disallow_empties)

        if repr is True:
            builder.add_repr(repr_ns)
        if str is True:
            builder.add_str()
        if cmp is True and not is_exc:
            builder.add_cmp()

        if hash is not True and hash is not False and hash is not None:
            # Can't use `hash in` because 1 == True for example.
            raise TypeError("Invalid value for hash.  Must be True, False, or None.")
        elif hash is False or (hash is None and cmp is False):
            if cache_hash:
                raise TypeError(
                    "Invalid value for cache_hash.  To use hash caching,"
                    " hashing must be either explicitly or implicitly "
                    "enabled."
                )
        elif (
            hash is True
            or (hash is None and cmp is True and frozen is True)
            and is_exc is False
        ):
            builder.add_hash()
        else:
            if cache_hash:
                raise TypeError(
                    "Invalid value for cache_hash.  To use hash caching,"
                    " hashing must be either explicitly or implicitly "
                    "enabled."
                )
            builder.make_unhashable()

        if init is True:
            builder.add_init()
        else:
            if cache_hash:
                raise TypeError(
                    "Invalid value for cache_hash.  To use hash caching,"
                    " init must be True."
                )

        return builder.build_class()

    # maybe_cls's type depends on the usage of the decorator.  It's a class
    # if it's used as `@attrs` but ``None`` if used as `@attrs()`.
    if maybe_cls is None:
        return wrap
    else:
        return wrap(maybe_cls)


def nonempty_validator(self, attribute, value):
    """Don't allow strings and collections without attr defaults to have empty/False-y values."""
    if attribute.type in _SCALAR_TYPES_WITH_NO_EMPTY_VALUES:
        # doesn't make sense to 'validate' these types against emptiness
        # since 0 or False are not 'empty' values
        return
    if attribute.default == NOTHING:
        if not value:
            raise ValueError(
                f'Attribute "{attribute.name}" on class {type(self)} '
                f"with type {attribute.type} "
                f"cannot have empty value '{value}'!"
            )


def _hack_add_validator(attrib, validator):
    """Hacks into an attrs attribute to validate something."""
    existing_validator = getattr(attrib, "validator")
    if existing_validator and existing_validator != nonempty_validator:
        input_validator = validator

        def combine_validators(self, att, val):
            existing_validator(self, att, val)
            input_validator(self, att, val)

        validator = combine_validators
    attrib._setattrs((("validator", validator),))


def _hook_builder_before_doing_anything(builder, disallow_empties=True):
    """This is a callback of sorts that I would love to be able
    to have in attrs by default."""
    if not disallow_empties:
        return  # no hook to run other than this one
    for attrib in builder._attrs:
        if attrib.type in _SCALAR_TYPES_WITH_NO_EMPTY_VALUES:
            continue
        # validate ALL attributes that don't have valid default values.
        if attrib.default == NOTHING:
            _hack_add_validator(attrib, nonempty_validator)


def get_attrs_names(Type: type) -> ty.Set[str]:
    attrs_attrs = getattr(Type, "__attrs_attrs__", None)
    if attrs_attrs is None:
        raise ValueError(f"type {Type} is not an attrs class")
    attrs: ty.Set[str] = {attr.name for attr in attrs_attrs}
    return attrs


def drop_nonattrs(d: dict, Type: type) -> dict:
    """gets rid of all members of the dictionary that wouldn't fit in the given 'attrs' Type"""
    attrs = get_attrs_names(Type)
    return {key: val for key, val in d.items() if key in attrs}



"""A Plugin for MyPy that helps it understand the Cats wrapper around attrs and cattrs"""
# flake8: noqa
# pylint: skip-file
# pylint: disable=import-error
import typing as ty
from mypy.nodes import (
    Var,
    Argument,
    ARG_POS,
    FuncDef,
    PassStmt,
    Block,
    SymbolTableNode,
    MDEF,
    SymbolTable,
    TypeInfo,
    FuncBase,
    SymbolNode,
)
from mypy.types import (
    NoneTyp,
    Type,
    CallableType,
    AnyType,
    TypeOfAny,
    Instance,
    UnionType,
)
from mypy.typevars import fill_typevars
from mypy.semanal import set_callable_name
from mypy.plugin import Plugin, ClassDefContext
from mypy.plugins.attrs import attr_class_maker_callback
from mypy.plugins.common import add_method


CAT_PATH = "typecats.tc.Cat"
STRUCTURE_NAME = "struc"
TRY_STRUCTURE_NAME = "try_struc"
UNSTRUCTURE_NAME = "unstruc"


def fullname(x: ty.Union[FuncBase, SymbolNode]) -> str:
    fn = x.fullname
    if callable(fn):
        return fn()
    return fn


def nameit(x: ty.Union[FuncBase, SymbolNode]) -> str:
    fn = x.name
    if callable(fn):
        return fn()
    return fn


def plugin(_version: str):
    """Plugin for MyPy Typechecking of Cats"""
    return CatsPlugin


def make_optional(typ: Type):
    return UnionType.make_union([typ, NoneTyp()])


class CatsPlugin(Plugin):
    """A plugin to make MyPy understand Cats"""

    def get_class_decorator_hook(
        self, fullname: str
    ) -> ty.Optional[ty.Callable[[ClassDefContext], None]]:
        """One of the MyPy Plugin defined entry points"""

        def add_struc_and_unstruc_to_classdefcontext(cls_def_ctx: ClassDefContext):
            """This MyPy hook tells MyPy that struc and unstruc will be present on a Cat"""

            dict_type = cls_def_ctx.api.named_type("__builtins__.dict")
            str_type = cls_def_ctx.api.named_type("__builtins__.str")
            api = cls_def_ctx.api
            implicit_any = AnyType(TypeOfAny.special_form)
            mapping = api.lookup_fully_qualified_or_none("typing.Mapping")
            if not mapping or not mapping.node:
                api.defer()
                return

            mapping_str_any_type = Instance(
                mapping.node,  # type: ignore # i don't understand this one but it works
                [str_type, implicit_any],
            )
            maybe_mapping_str_any_type = make_optional(mapping_str_any_type)

            if fullname == CAT_PATH:
                attr_class_maker_callback(
                    cls_def_ctx, True
                )  # since a Cat is also an attr.s class...
                info = cls_def_ctx.cls.info
                cat_return_type = fill_typevars(info)
                maybe_cat_return_type = make_optional(cat_return_type)

                if STRUCTURE_NAME not in info.names:
                    add_static_method(
                        cls_def_ctx,
                        STRUCTURE_NAME,
                        [
                            Argument(
                                Var("d", mapping_str_any_type),
                                mapping_str_any_type,
                                None,
                                ARG_POS,
                            )
                        ],
                        cat_return_type,
                    )
                if TRY_STRUCTURE_NAME not in info.names:
                    add_static_method(
                        cls_def_ctx,
                        TRY_STRUCTURE_NAME,
                        [
                            Argument(
                                Var("d", maybe_mapping_str_any_type),
                                maybe_mapping_str_any_type,
                                None,
                                ARG_POS,
                            )
                        ],
                        maybe_cat_return_type,
                    )
                if UNSTRUCTURE_NAME not in info.names:
                    add_method(cls_def_ctx, UNSTRUCTURE_NAME, [], dict_type)

        if fullname == CAT_PATH:
            return add_struc_and_unstruc_to_classdefcontext
        return None


def serialize_info_name(
    info: TypeInfo, name: str, class_path: str
) -> ty.Dict[str, ty.Any]:
    slzed = info.names[name].serialize(class_path, name)
    return slzed


def deserialize_funcdefs(stmts):
    return [FuncDef.deserialize(stmt) for stmt in stmts if stmt[".class"] == "FuncDef"]


def add_static_method(
    ctx, function_name: str, args: ty.List[Argument], return_type: Type
) -> None:
    """Mostly copied from mypy.plugins.common, with changes to make it work for a static method."""
    info = ctx.cls.info
    function_type = ctx.api.named_type("__builtins__.function")

    arg_types, arg_names, arg_kinds = [], [], []
    for arg in args:
        assert arg.type_annotation, "All arguments must be fully typed."
        arg_types.append(arg.type_annotation)
        arg_names.append(nameit(arg.variable))
        arg_kinds.append(arg.kind)

    signature = CallableType(
        arg_types, arg_kinds, arg_names, return_type, function_type
    )

    func = FuncDef(function_name, args, Block([PassStmt()]))
    func.is_static = True
    func.info = info
    func.type = set_callable_name(signature, func)
    func._fullname = fullname(info) + "." + function_name
    func.line = info.line

    info.names[function_name] = SymbolTableNode(MDEF, func, plugin_generated=True)
    info.defn.defs.body.append(func)



"""Utilities for using attrs types with cattrs"""

import typing as ty
import traceback
import logging
from functools import partial

import cattr

from .attrs_shim import cat_attrs, get_attrs_names
from .wildcat import (
    WC,
    MWC,
    mixin_wildcat_post_attrs_methods,
    setup_warnings_for_dangerous_dict_subclass_operations,
)


logger = logging.getLogger(__name__)


CommonStructuringExceptions = (TypeError, AttributeError, ValueError)


C = ty.TypeVar("C")

StrucInput = ty.Mapping[str, ty.Any]
UnstrucOutput = dict


class TypeCat:
    """This is mostly just an example of the interface - it is unused
    within typecats itself.

    You could use this base class to make PyLint happier with the Cat
    decorator, which it doesn't understand. It's probably cleaner to
    just tell Pylint not to worry about it, by ignoring generated
    methods with these names.
    """

    @staticmethod
    def struc(_d: StrucInput) -> ty.Any:
        return TypeCat()  # this is a lie - don't worry about it

    @staticmethod
    def try_struc(_d: ty.Optional[StrucInput]) -> ty.Any:
        return TypeCat()

    def unstruc(self) -> UnstrucOutput:
        return unstruc(self)


TYPECATS_CONVERTER = cattr.Converter()


def struc(cl: ty.Type[C], obj: StrucInput) -> C:
    """A wrapper for cattrs structure that logs and re-raises structure exceptions."""
    try:
        return TYPECATS_CONVERTER.structure(obj, cl)
    except CommonStructuringExceptions as e:
        _log_structure_exception(e, obj, cl)
        raise e


def unstruc(obj: ty.Any) -> UnstrucOutput:
    """A wrapper for cattrs unstructure using the internal converter"""
    return TYPECATS_CONVERTER.unstructure(obj)


def struc_wildcat(Type: ty.Type[MWC], d: ty.Mapping) -> MWC:
    """A Wildcat is a Cat (an attrs class) that additionally allows
    arbitrary key-value access as though it were a dict for data that
    does not have a defined attribute name on the class (i.e. is not typed).

    This is intended to provide a smooth transition for types that
    have some dynamic elements, but as though dynamic elements become
    'settled', they can be directly typed and statically
    typed-checked. It also makes for a good partially-typed
    passthrough defintion if you need to retain and later transmit
    unknown keys while still letting your code reason about the types
    that you do know about.

    """
    try:
        wildcat_attrs_names = get_attrs_names(Type)
        # use our converter directly to avoid infinite recursion
        wildcat = TYPECATS_CONVERTER.structure_attrs_fromdict(d, Type)  # type: ignore
        # we have a partial wildcat. now add in all the things that weren't structured
        wildcat.update({key: d[key] for key in d if key not in wildcat_attrs_names})
        return wildcat
    except CommonStructuringExceptions as e:
        _log_structure_exception(e, d, Type)
        raise e


def unstruc_wildcat(wildcat: WC) -> UnstrucOutput:
    """Unstructures a Wildcat by extracting the untyped key/value pairs,

    then updating that dict with the result of the typed attrs object unstructure.

    Note that this always chooses the typed value in any key collisions if the Wildcat
    implementation happens to allow those.
    """
    wildcat_attrs_names = get_attrs_names(type(wildcat))
    # use our converter directly to avoid infinite recursion
    wildcat_dict = TYPECATS_CONVERTER.unstructure_attrs_asdict(wildcat)  # type: ignore
    wildcat_nonattrs_dict = {
        key: unstruc(wildcat[key]) for key in wildcat if key not in wildcat_attrs_names
    }
    # note that typed entries take absolute precedence over untyped in case of collisions.
    # these collisions should generally be prevented at runtime by the wildcat
    # logic that is injected into the type, but if something were to sneak through
    # we would prefer whatever had been set via the attribute.
    return {**wildcat_nonattrs_dict, **wildcat_dict}


def make_try_struc(
    structure_method: ty.Callable[[ty.Type[C], StrucInput], C],
    cl: ty.Type[C],
    obj: ty.Optional[StrucInput],
) -> ty.Optional[C]:
    """A wrapper for cattrs structure that suppresses and logs structure exceptions."""
    try:
        return structure_method(cl, obj)  # type: ignore
    except CommonStructuringExceptions:
        return None
    except Exception as e:
        _log_structure_exception(e, obj, cl)
        return None


try_struc = partial(make_try_struc, struc)
try_struc_wildcat = partial(make_try_struc, struc_wildcat)


def register_struc_hook(*args, **kwargs):
    """Use this to register cattrs structuring hooks on the internal cattrs Converter"""
    TYPECATS_CONVERTER.register_structure_hook(*args, **kwargs)


def register_unstruc_hook(*args, **kwargs):
    """Use this to register cattrs unstructuring hooks on the internal cattrs Converter"""
    TYPECATS_CONVERTER.register_unstructure_hook(*args, **kwargs)


STRUCTURE_NAME = "struc"
TRY_STRUCTURE_NAME = "try_struc"
UNSTRUCTURE_NAME = "unstruc"


def Cat(maybe_cls=None, auto_attribs=True, disallow_empties=True, **kwargs):
    """A Cat knows how to take care of itself.

    This decorator combines the beauty of attrs-style classes with the
    extreme, ferocious power of cattrs, and provides it all in a convenient package.
    Simply import your type and call YourType.struc(dict_of_your_type), and
    you'll have an object of your type returned to you! Want a raw
    dict back? YourObject.unstruc()!

    Sublime.

    Additionally: Cats don't like to feel empty. If you've defined an
    attribute without a default value, it is assumed that a non-empty
    value is required for that attribute, and a validator requiring
    that will automatically be added. If you for some reason want to
    require attributes but allow empty values (very strange!), you can turn
    off this behavior with the pedestrianly-named boolean flag
    'disallow_empties=False'.

    """

    def make_cat(cls: ty.Type[C]) -> ty.Type[C]:
        is_wild = dict in cls.__mro__
        if is_wild:
            setup_warnings_for_dangerous_dict_subclass_operations(cls)
            register_struc_hook(cls, lambda obj, typ: struc_wildcat(typ, obj))
            register_unstruc_hook(cls, unstruc_wildcat)

        # it is always safe to apply this decorator, even if there's already an attrs_attrs on a base class.
        cls = cat_attrs(
            cls, auto_attribs=auto_attribs, disallow_empties=disallow_empties, **kwargs
        )

        cat_structure_method = struc_wildcat if is_wild else struc
        cat_try_struc = partial(make_try_struc, cat_structure_method)
        cat_unstructure_method = unstruc_wildcat if is_wild else unstruc

        @staticmethod  # type: ignore
        def struc_cat(d: StrucInput) -> C:
            return cat_structure_method(cls, d)

        setattr(cls, STRUCTURE_NAME, struc_cat)

        @staticmethod  # type: ignore
        def try_struc_cat(d: ty.Optional[StrucInput]) -> ty.Optional[C]:
            return cat_try_struc(cls, d)

        setattr(cls, TRY_STRUCTURE_NAME, try_struc_cat)

        def unstruc_to_dict(self) -> UnstrucOutput:
            return cat_unstructure_method(self)

        setattr(cls, UNSTRUCTURE_NAME, unstruc_to_dict)

        if is_wild:
            mixin_wildcat_post_attrs_methods(cls)

        return cls

    if maybe_cls is None:
        return make_cat
    return make_cat(maybe_cls)


def _log_structure_exception(exception: Exception, item: ty.Any, Type: type):
    type_name = getattr(Type, "__name__", str(Type))
    logger.info(
        f"Failed to structure {type_name} from {item}",
        extra=dict(
            json=dict(item=item),
            traceback=traceback.format_exception(
                None, exception, exception.__traceback__
            ),
        ),
    )



"""Implements some nice-to-haves for the 'Wildcat' concept of structurable dynamic Class properties."""
# pylint: disable=protected-access
import typing as ty
import logging

from .attrs_shim import get_attrs_names


logger = logging.getLogger(__name__)


MWC = ty.TypeVar("MWC", bound=ty.MutableMapping)
WC = ty.TypeVar("WC", bound=ty.Mapping)


def is_wildcat(cls_or_obj: ty.Union[object, type]) -> bool:
    return (
        hasattr(cls_or_obj, "__setitem__")
        and hasattr(cls_or_obj, "__getitem__")
        and hasattr(cls_or_obj, "__iter__")
    )


def _strip_defined_abstract_methods(cls):
    """If a method has been dynamically defined/mixed-in, then it is no longer abstract,

    but apparently this must be fixed up manually
    """
    abs_methods = getattr(cls, "__abstractmethods__", None)
    if abs_methods:
        new_abs_methods = {
            methodname for methodname in abs_methods if methodname not in cls.__dict__
        }
        setattr(cls, "__abstractmethods__", frozenset(new_abs_methods))


def setup_warnings_for_dangerous_dict_subclass_operations(cls):
    """Adds safeguards that will warn about attributes that 'overlap' keys
    to a class that inherits from dict.
    """
    class_name = getattr(cls, "__name__", str(cls))

    def warn_key_set_on_attribute(key):
        logger.warning(
            f"Attribute '{key}' is explicitly typed on '{class_name}' "
            "so this should be changed to attribute assigment."
        )

    def __setitem__(self, key, item):
        if hasattr(self, key):
            warn_key_set_on_attribute(key)
            setattr(self, key, item)
        else:
            super(cls, self).__setitem__(key, item)  # type: ignore

    setattr(cls, "__setitem__", __setitem__)

    def __getitem__(self, key):
        if hasattr(self, key):
            logger.warning(
                f"Attribute '{key}' is explicitly typed on '{class_name}' "
                "so this should be changed to attribute access."
            )
            return getattr(self, key)
        return super(cls, self).__getitem__(key)  # type: ignore

    setattr(cls, "__getitem__", __getitem__)

    def update(self, other_dict=None, **kwargs):
        if not other_dict:
            other_dict = kwargs
        non_attribute_kvs = {
            k: v for k, v in other_dict.items() if not hasattr(self, k)
        }
        for key, value in other_dict.items():
            if key not in non_attribute_kvs:
                self[key] = value  # reuse __setitem__ which will forward to setattr
        super(cls, self).update(non_attribute_kvs)  # type: ignore

    setattr(cls, "update", update)


def mixin_wildcat_post_attrs_methods(cls):
    """Adds a repr to an attrs class that additionally prints out an internal dict"""

    def __repr__(self):
        if dict in cls.__mro__:
            wildcat_part = f"+Wildcat{dict(self.items())}" if self.keys() else ""
        else:
            wildcat_part = (
                f"+Wildcat{self.__wildcat_dict}" if self.__wildcat_dict else ""
            )
        return self.__attrs_repr__() + wildcat_part

    setattr(cls, "__attrs_repr__", cls.__repr__)
    setattr(cls, "__repr__", __repr__)

    # we ran into an issue where, because a Wildcat is a dict and therefore defined
    # __len__, that value was used as the truth-check in a standard `if wildcat:`
    # check, which was not what was desired since the object itself existed and had data.
    def __bool__(self):
        """An actual Wildcat is truthy based on the entire contents of its
        attributes and dict, unless otherwise defined"""
        is_truthy = bool(len(self))
        for attr_name in get_attrs_names(cls):
            if is_truthy:
                break
            is_truthy |= bool(getattr(self, attr_name, False))
        return is_truthy

    if not hasattr(cls, "__bool__"):
        setattr(cls, "__bool__", __bool__)

    _strip_defined_abstract_methods(cls)



"""Setup for typecats"""
# type: ignore
from setuptools import setup, find_packages

PKG_NAME = "typecats"
about = {}  # type: ignore
exec(open(f"{PKG_NAME}/__about__.py").read(), about)  # pylint: disable=exec-used

with open("README.md") as fh:
    long_description = fh.read()

setup(
    name=PKG_NAME,
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    url="https://github.com/xoeye/typecats",
    description="Structure unstructured data for the purpose of static type checking",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    package_data={"": ["py.typed"]},
    python_requires=">=3.6",
    install_requires=[
        "attrs == 19.1",
        "cattrs == 1.0.0rc0",
        # 'python-dateutil ~= 2.8',
    ],
    # it is important to keep these install_requires basically in sync with the Pipfile as well.
)
